/*
 * Hybrid Logic Model Checker 0.9 - HL Model I/O functions and data structures
 * 
 * Copyright (C) 2005 Luigi Dragone
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; version 2.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *   
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Luigi Dragone <luigi@luigidragone.com>
 */


#include "linked-list.h"
#include "hl-model.h"
#include <stdio.h>
#include <expat.h>
#include <string.h>
#include <errno.h>

// XML element and attribute names and default values
#define HL_KS_ELEM "hl-kripke-struct"
#define HL_KS_NAME_ATTR "name"
#define HL_KS_NAME_ATTR_DEF "M"
#define WORLD_ELEM "world"
#define WORLD_LABEL_ATTR "label"
#define MODALITY_ELEM "modality"
#define MODALITY_LABEL_ATTR "label"
#define ACC_PAIR_ELEM "acc-pair"
#define ACC_PAIR_FROM_WORLD_LABEL_ATTR "from-world-label"
#define ACC_PAIR_TO_WORLD_LABEL_ATTR "to-world-label"
#define PROP_SYM_ELEM "prop-sym"
#define PROP_SYM_LABEL_ATTR "label"
#define PROP_SYM_TRUTH_ASSIGNMENTS_ATTR "truth-assignments"
#define NOMINAL_ELEM "nominal"
#define NOMINAL_LABEL_ATTR "label"
#define NOMINAL_TRUTH_ASSIGNMENT_ATTR "truth-assignment"
#define IDREFS_DELIM " "

/***
 * Temporany modality access relation matrix used during parsing
 */
typedef struct ACC_REL_MATRIX_s {
	unsigned int identifier;
	HL_BOOLEAN** matrix;
	unsigned int size;
} ///
ACC_REL_MATRIX_t;

/***
 * Type constructor
 */
static ACC_REL_MATRIX_t* ACC_REL_MATRIX_new(unsigned int identifier, unsigned int size) {
	int i, j;
	ACC_REL_MATRIX_t* res = malloc(sizeof(ACC_REL_MATRIX_t));
	res->size = size;
	res->identifier = identifier;
	res->matrix = malloc(sizeof(HL_BOOLEAN*) * size);
	for(i = 0; i < size; i++) {
		res->matrix[i] = malloc(sizeof(HL_BOOLEAN) * size);
		for(j = 0; j < size; j++)
			res->matrix[i][j] = HL_FALSE;	
	}
	return res;
}

/***
 * Type destructor
 */
static void ACC_REL_MATRIX_delete(ACC_REL_MATRIX_t* a) {
	int i;
	if(a != NULL) {
		for(i = 0; i < a->size; i++)
			free(a->matrix[i]);
		free(a->matrix);
		free(a);
	}
}

/***
 * Temporany propositional symbol truth assignment used during parsing
 */
typedef struct TRUTH_ASSIGNMENT_s {
	unsigned int identifier;
	char* worldLabels;
} ///
TRUTH_ASSIGNMENT_t;

/***
 * Type constructor
 */
static TRUTH_ASSIGNMENT_t* TRUTH_ASSIGNMENT_new(unsigned int identifier, const char* worldLabels) {
	TRUTH_ASSIGNMENT_t* a = malloc(sizeof(TRUTH_ASSIGNMENT_t));
	a->identifier = identifier;
	a->worldLabels = strdup(worldLabels);
	return a;
} 

/***
 * Type destructor
 */
static void TRUTH_ASSIGNMENT_delete(TRUTH_ASSIGNMENT_t* a) {
	if(a != NULL) {
		free(a->worldLabels);
		free(a);
	}
}

/***
 * Expat parser callback function data.
 *
 * This structure describes data manipulated across different 
 * callback functions during the parsing of XML file, please refer
 * to Expat documentation for further details about how the parser works.
 */
typedef struct MODEL_PARSER_DATA_s {
	KRIPKE_STRUCT_t* theModel;
	const char* modelName;
	HL_BOOLEAN errorStatus;
	unsigned int worldCount;
	unsigned int modalityCount;
	unsigned int propSymCount;
	unsigned int nominalCount;
	LINKED_LIST_t* nominalAssigns;
	LINKED_LIST_t* propSymAssigns;
	LINKED_LIST_t* modAccRels;
	SYMBOL_TABLE_t* propSymTable;
	SYMBOL_TABLE_t* nominalTable;
	SYMBOL_TABLE_t* modalityTable;
	SYMBOL_TABLE_t* worldTable;
  	XML_Parser parser;
}  MODEL_PARSER_DATA_t;

/***
 * Type constructor
 */
static MODEL_PARSER_DATA_t* MODEL_PARSER_DATA_new(XML_Parser parser) {
	MODEL_PARSER_DATA_t* res = malloc(sizeof(MODEL_PARSER_DATA_t));
	res->errorStatus = HL_FALSE;
	res->theModel = NULL;
	res->worldCount = 0;
	res->modalityCount = 0;
	res->propSymCount = 0;
	res->nominalCount = 0;
	res->modAccRels = NULL;
	res->propSymAssigns = NULL;
	res->nominalAssigns = NULL;
	res->parser = parser;
	return res;
}

/***
 * Type destructor
 */
static void MODEL_PARSER_DATA_delete(MODEL_PARSER_DATA_t* p) {
	if(p != NULL) {
		LINKED_LIST_t* listPtr = p->modAccRels;
		while(listPtr != NULL) {
			ACC_REL_MATRIX_delete(listPtr->data);
			listPtr = listPtr->next;
		}
		listPtr = p->nominalAssigns;
		while(listPtr != NULL) {
			TRUTH_ASSIGNMENT_delete(listPtr->data);
			listPtr = listPtr->next;
		}
		listPtr = p->propSymAssigns;
		while(listPtr != NULL) {
			TRUTH_ASSIGNMENT_delete(listPtr->data);
			listPtr = listPtr->next;
		}
		LINKED_LIST_delete(p->nominalAssigns);
		LINKED_LIST_delete(p->propSymAssigns);
		LINKED_LIST_delete(p->modAccRels);
		free(p);	
	}
}

/***
 * Utility function to look up for a specific attribute value in Expat data structure
 */
static const char* attrValueLookup(const char **attrs, const char* attrName) {
	int i;
  	for (i = 0; attrs[i]; i += 2)
 		if(!strcmp(attrName, attrs[i])) 
  			return attrs[i + 1];
  	return NULL;
}

/***
 * Error reporting function
 */
static void reportMissingAttributeError(MODEL_PARSER_DATA_t* parserData, const char* attrName) {
	fprintf(stderr, 
    		"Missing attribute %s at line %d\n",
		attrName,
		(int)XML_GetCurrentLineNumber(parserData->parser));
}

/***
 * Error reporting function
 */
static void reportInvalidAttributeError(MODEL_PARSER_DATA_t* parserData, const char* attrName, const char* attrValue) {
	fprintf(stderr, 
    		"Invalid value \'%s\' for attribute %s at line %d\n",
	    	attrValue,
		attrName,
		(int)XML_GetCurrentLineNumber(parserData->parser));
}

/***
 * Error reporting function
 */
static void reportInvalidNominalAssignmentError(MODEL_PARSER_DATA_t* parserData, const char* nominalLabel, const char* assignment) {
	fprintf(stderr, 
    		"Invalid assignment \'%s\' for nominal %s\n",
    		assignment,
	    	nominalLabel);
}

/***
 * Error reporting function
 */
static void reportInvalidPropSymAssignmentError(MODEL_PARSER_DATA_t* parserData, const char* propSymLabel, const char* assignment) {
	fprintf(stderr, 
    	"Invalid assignment \'%s\' for propositional symbol %s\n",
    	assignment,
	    propSymLabel);
}

/***
 * Expat callback function
 */
static void XMLCALL startElement(void *userData, const char *name, const char **attrs) {
	int i;
	MODEL_PARSER_DATA_t* parserData = userData;
  	if(!strcmp(HL_KS_ELEM, name)) {
  		parserData->modelName = strdup(attrValueLookup(attrs, HL_KS_NAME_ATTR));
  		if(parserData->modelName == NULL)
  			parserData->modelName = HL_KS_NAME_ATTR_DEF;
  	} else if(!strcmp(WORLD_ELEM, name)) {
  		const char* worldLabel = attrValueLookup(attrs, WORLD_LABEL_ATTR);
  		if(worldLabel != NULL) {
  			SYMBOL_TABLE_addLabel(parserData->worldTable, worldLabel);	
  			parserData->worldCount++;
  		} else {
  			parserData->errorStatus = HL_TRUE;
  			reportMissingAttributeError(parserData, WORLD_LABEL_ATTR);
  		}
  	} else if(!strcmp(MODALITY_ELEM, name)) {
  		const char* modalityLabel = attrValueLookup(attrs, MODALITY_LABEL_ATTR);
  		if(modalityLabel != NULL) {
  			unsigned int id = SYMBOL_TABLE_addLabel(parserData->modalityTable, modalityLabel);	
  			parserData->modalityCount++;
			ACC_REL_MATRIX_t* accRel = ACC_REL_MATRIX_new(id, parserData->worldCount);
			parserData->modAccRels = LINKED_LIST_new(accRel, parserData->modAccRels);
  		} else {
  			parserData->errorStatus = HL_TRUE;
   			reportMissingAttributeError(parserData, MODALITY_LABEL_ATTR);
  		}
   	} else if(!strcmp(ACC_PAIR_ELEM, name)) {
   		const char* fromWorldLabel = NULL;
   		const char* toWorldLabel = NULL;
  		for (i = 0; attrs[i]; i += 2)
	 		if(!strcmp(ACC_PAIR_FROM_WORLD_LABEL_ATTR, attrs[i])) 
	 			fromWorldLabel = attrs[i + 1];
	 		else if(!strcmp(ACC_PAIR_TO_WORLD_LABEL_ATTR, attrs[i])) 
	 			toWorldLabel = attrs[i + 1];
	 	if(fromWorldLabel != NULL && toWorldLabel != NULL) {
	 		unsigned int fromId = SYMBOL_TABLE_getLabelIdentifier(parserData->worldTable, fromWorldLabel);	
	 		unsigned int toId = SYMBOL_TABLE_getLabelIdentifier(parserData->worldTable, toWorldLabel);	
	 		if(fromId != HL_INVALID_LABEL && toId != HL_INVALID_LABEL) {
	 			ACC_REL_MATRIX_t* accRel = parserData->modAccRels->data;
	 			accRel->matrix[fromId][toId] = HL_TRUE;
	 		} else {
		 		if(fromId == HL_INVALID_LABEL) 
	 				reportInvalidAttributeError(parserData, ACC_PAIR_FROM_WORLD_LABEL_ATTR, fromWorldLabel);
	 			if(toId == HL_INVALID_LABEL)
		 			reportInvalidAttributeError(parserData, ACC_PAIR_TO_WORLD_LABEL_ATTR, toWorldLabel);
  				parserData->errorStatus = HL_TRUE;
  			}
	 	} else {
	 		if(fromWorldLabel == NULL)
	  			reportMissingAttributeError(parserData, ACC_PAIR_FROM_WORLD_LABEL_ATTR);
	  		if(toWorldLabel == NULL)
	  			reportMissingAttributeError(parserData, ACC_PAIR_TO_WORLD_LABEL_ATTR);
  			parserData->errorStatus = HL_TRUE;
  		}
  	} else if(!strcmp(PROP_SYM_ELEM, name)) {
  		const char* propSymLabel = NULL;
   		const char* assignments = NULL;
  		for (i = 0; attrs[i]; i += 2)
	 		if(!strcmp(PROP_SYM_LABEL_ATTR, attrs[i])) 
	 			propSymLabel = attrs[i + 1];
	 		else if(!strcmp(PROP_SYM_TRUTH_ASSIGNMENTS_ATTR, attrs[i])) 
	 			assignments = attrs[i + 1];
	 	if(propSymLabel != NULL) {
	 		unsigned int id = SYMBOL_TABLE_addLabel(parserData->propSymTable, propSymLabel);	
	  		parserData->propSymCount++;
			TRUTH_ASSIGNMENT_t* propSymAssigns = TRUTH_ASSIGNMENT_new(id, assignments);
			parserData->propSymAssigns = LINKED_LIST_new(propSymAssigns, parserData->propSymAssigns);
		} else {
  			reportMissingAttributeError(parserData, PROP_SYM_LABEL_ATTR);
  			parserData->errorStatus = HL_TRUE;
  		}
  	} else if(!strcmp(NOMINAL_ELEM, name)) {
  		const char* nominalLabel = NULL;
   		const char* assignment = NULL;
  		for (i = 0; attrs[i]; i += 2)
	 		if(!strcmp(NOMINAL_LABEL_ATTR, attrs[i])) 
	 			nominalLabel = attrs[i + 1];
	 		else if(!strcmp(NOMINAL_TRUTH_ASSIGNMENT_ATTR, attrs[i])) 
	 			assignment = attrs[i + 1];
	 	if(nominalLabel != NULL && assignment != NULL) {
	 		unsigned int id = SYMBOL_TABLE_addLabel(parserData->nominalTable, nominalLabel);	
	  		parserData->nominalCount++;
			TRUTH_ASSIGNMENT_t* nominalAssigns = TRUTH_ASSIGNMENT_new(id, assignment);
			parserData->nominalAssigns = LINKED_LIST_new(nominalAssigns, parserData->nominalAssigns);
	 	} else {
	 		if(nominalLabel == NULL)
	  			reportMissingAttributeError(parserData, NOMINAL_LABEL_ATTR);
	  		if(assignment == NULL)
		  		reportMissingAttributeError(parserData, NOMINAL_TRUTH_ASSIGNMENT_ATTR);
  			parserData->errorStatus = HL_TRUE;
  		}
  	} 
}

/***
 * Expat callback function
 */
static void XMLCALL endElement(void *userData, const char *name) {
	MODEL_PARSER_DATA_t* parserData = userData;
	if(!strcmp(HL_KS_ELEM, name) && !parserData->errorStatus) {
		unsigned int i, j;
  		parserData->theModel = KRIPKE_STRUCT_new(parserData->worldCount, 
	  		parserData->modalityCount, parserData->propSymCount, parserData->nominalCount);
  		LINKED_LIST_t* listPtr = NULL;
  		ACC_REL_MATRIX_t* accRel = NULL;
		TRUTH_ASSIGNMENT_t* assignment = NULL;
		unsigned int nominalId;
		unsigned int worldId;
		unsigned int propSymId;
		char* worldLabels;
		char* worldLabel;
  		for(listPtr = parserData->modAccRels; listPtr != NULL; listPtr = listPtr->next) {
			accRel = listPtr->data;
			for(i = 0; i < parserData->worldCount; i++)
				for(j = 0; j < parserData->worldCount; j++)
					KRIPKE_STRUCT_setAccessibile(parserData->theModel, 
					accRel->identifier, i, j, accRel->matrix[i][j]);			
		}
  		for(listPtr = parserData->nominalAssigns; listPtr != NULL; listPtr = listPtr->next) {
  			assignment = listPtr->data;
  			nominalId = assignment->identifier;
//  			if(assignment->worldLabels != NULL) {
	  			worldLabel = assignment->worldLabels;
  				worldId = SYMBOL_TABLE_getLabelIdentifier(parserData->worldTable, worldLabel);
  				if(worldId != HL_INVALID_LABEL)
  					KRIPKE_STRUCT_setNominalAssign(parserData->theModel, nominalId, worldId);
  				else {
					reportInvalidNominalAssignmentError(parserData, NULL, worldLabel);
  					parserData->errorStatus = HL_TRUE;
  				}
  //			}
  		}
  		for(listPtr = parserData->propSymAssigns; listPtr != NULL; listPtr = listPtr->next) {
  			assignment = listPtr->data;
  			propSymId = assignment->identifier;
  			if(assignment->worldLabels != NULL) {
	  			worldLabels = assignment->worldLabels;
  				for(worldLabel = strtok(worldLabels, IDREFS_DELIM); worldLabel != NULL; worldLabel = strtok(NULL, IDREFS_DELIM)) {
  					worldId = SYMBOL_TABLE_getLabelIdentifier(parserData->worldTable, worldLabel);
  					if(worldId != HL_INVALID_LABEL)
		  				KRIPKE_STRUCT_setPropSym(parserData->theModel, propSymId, worldId, HL_TRUE); 
		  			else {
						reportInvalidPropSymAssignmentError(parserData, NULL, worldLabel);
  						parserData->errorStatus = HL_TRUE;
  					}
  				} 
  			}
  		}
  	} 	
}

HL_MODEL_t* HL_MODEL_new(const char *fileName) {
	FILE* file;
  	char buf[BUFSIZ];
  	HL_MODEL_t* model = NULL;
  	int done;
  	XML_Parser parser = NULL;
  	MODEL_PARSER_DATA_t* parserData = NULL;
  	
  	file = fopen(fileName, "r");
  	if(file == NULL) {
		fprintf (stderr, "Couldn't open input file \'%s\': %s\n", fileName, strerror(errno));
  		return model;
  	}
  	parser = XML_ParserCreate(NULL);
	parserData = MODEL_PARSER_DATA_new(parser);
	  	
	parserData->propSymTable = SYMBOL_TABLE_new();
	parserData->nominalTable = SYMBOL_TABLE_new();
	parserData->modalityTable = SYMBOL_TABLE_new();
	parserData->worldTable = SYMBOL_TABLE_new();

  	XML_SetUserData(parser, parserData);
  	XML_SetElementHandler(parser, startElement, endElement);
  	do {
    	size_t len = fread(buf, 1, sizeof(buf), file);
    	done = len < sizeof(buf);
    	if (XML_Parse(parser, buf, len, done) == XML_STATUS_ERROR) {
      		fprintf(stderr,
        		"%s at line %d\n",
	        	XML_ErrorString(XML_GetErrorCode(parser)),
    	        	(int)XML_GetCurrentLineNumber(parser));
      		fclose(file);
			return model;
    	}
  	} while (!done);
  	model = malloc(sizeof(HL_MODEL_t));
  	model->kripkeStruct = parserData->theModel;
  	
  	model->propSymTable = parserData->propSymTable;
	model->nominalTable = parserData->nominalTable;
	model->modalityTable = parserData->modalityTable;
	model->worldTable = parserData->worldTable;
  	model->modelName = parserData->modelName;
  	MODEL_PARSER_DATA_delete(parserData);
  	XML_ParserFree(parser);
  	fclose(file);
  	return model;
}

void HL_MODEL_printKripkeStruct(FILE* os, HL_MODEL_t* model) {
	if(model != NULL) {
		int i, j, k;
		KRIPKE_STRUCT_t* ks = model->kripkeStruct;
		SYMBOL_ARRAY_t* propSymLabels = SYMBOL_ARRAY_new(model->propSymTable);
		SYMBOL_ARRAY_t* nominalLabels = SYMBOL_ARRAY_new(model->nominalTable);
		SYMBOL_ARRAY_t* modalityLabels = SYMBOL_ARRAY_new(model->modalityTable);
		SYMBOL_ARRAY_t* worldLabels = SYMBOL_ARRAY_new(model->worldTable);
		fprintf(os, "Kripke structure: %s\n", model->modelName);
		fprintf(os, "\tWorlds: ");
		for(i = 0; i < ks->worldNum; i++)
			fprintf(os, "%s (%d)%s", SYMBOL_ARRAY_getEntry(worldLabels, i), i, i != ks->worldNum - 1 ? ", " : "\n");
		fprintf(os, "\tModalities:\n");
		for(i = 0; i < ks->modNum; i++) {
			fprintf(os, "\t\t%s (%d) = ", SYMBOL_ARRAY_getEntry(modalityLabels, i), i);
			for(j = 0; j < ks->worldNum; j++)
				for(k = 0; k < ks->worldNum; k++)
					if(ks->accessibilityRels[i][j][k])
						fprintf(os, "<%s, %s> ", SYMBOL_ARRAY_getEntry(worldLabels, j), SYMBOL_ARRAY_getEntry(worldLabels, k));
			fprintf(os, "\n");
						
		}
		fprintf(os, "\tPropositional symbols:\n");
		for(i = 0; i < ks->propSymNum; i++) {
			fprintf(os, "\t\t%s (%d) = ", SYMBOL_ARRAY_getEntry(propSymLabels, i), i);
			for(j = 0; j < ks->worldNum; j++)
				if(ks->propSymAssigns[i][j])
					fprintf(os, "%s ", SYMBOL_ARRAY_getEntry(worldLabels, j));
			fprintf(os, "\n");				
		}
		fprintf(os, "\tNominals:\n");
		for(i = 0; i < ks->nominalNum; i++) {
			fprintf(os, "\t\t%s (%d) = %s\n", SYMBOL_ARRAY_getEntry(nominalLabels, i), i, SYMBOL_ARRAY_getEntry(worldLabels, ks->nominalAssigns[i]));
		}
		SYMBOL_ARRAY_delete(worldLabels);
		SYMBOL_ARRAY_delete(propSymLabels);
		SYMBOL_ARRAY_delete(nominalLabels);
		SYMBOL_ARRAY_delete(modalityLabels);
	}	
	
}

void HL_MODEL_printFormula(FILE* os, HL_MODEL_t* model, HL_FORMULA_t* formula, SYMBOL_TABLE_t* wvarTable) {
	if(model != NULL && formula != NULL) {
		SYMBOL_ARRAY_t* propSymLabels = SYMBOL_ARRAY_new(model->propSymTable);
		SYMBOL_ARRAY_t* nominalLabels = SYMBOL_ARRAY_new(model->nominalTable);
		SYMBOL_ARRAY_t* modalityLabels = SYMBOL_ARRAY_new(model->modalityTable);
		SYMBOL_ARRAY_t* wvarLabels = SYMBOL_ARRAY_new(wvarTable);
		HL_FORMULA_prettyPrint(os, formula, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
		SYMBOL_ARRAY_delete(propSymLabels);
		SYMBOL_ARRAY_delete(nominalLabels);
		SYMBOL_ARRAY_delete(modalityLabels);
		SYMBOL_ARRAY_delete(wvarLabels);
	}
}

void HL_MODEL_printMCResults(FILE* os, HL_MODEL_t* model, HL_FORMULA_t* formula) {
	if(model != NULL && formula != NULL) {
		int i;
		SYMBOL_ARRAY_t* worldLabels = SYMBOL_ARRAY_new(model->worldTable);
		for(i = 0; i < formula->model->worldNum; i++)
			if(formula->worldEvals[i])
				fprintf(os, "%s ", SYMBOL_ARRAY_getEntry(worldLabels, i));
		SYMBOL_ARRAY_delete(worldLabels);	
	}
}

void HL_MODEL_delete(HL_MODEL_t* model) {
	if(model != NULL) {
		KRIPKE_STRUCT_delete(model->kripkeStruct);
		SYMBOL_TABLE_delete(model->propSymTable);	
		SYMBOL_TABLE_delete(model->nominalTable);
		SYMBOL_TABLE_delete(model->modalityTable);
		SYMBOL_TABLE_delete(model->worldTable);		
		if(strcmp(model->modelName, HL_KS_NAME_ATTR_DEF))
			free((void*)model->modelName);
		free(model);
	}	
}
