/*
 * Hybrid Logic Model Checker 0.9 - HL Model I/O functions and data structures header file
 * 
 * Copyright (C) 2005 Luigi Dragone
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; version 2.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *   
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Luigi Dragone <luigi@luigidragone.com>
 */

#ifndef __HL_MODEL_H__
#define __HL_MODEL_H__

/** Hybrid logic functions and data types */
#include "hl.h"

/**
 * Model data type for HL.
 * 
 * The model is composed by the Kripke structure and the symbol tables of 
 * propositional symbols, modalities, nominals and worlds. It can be identified
 * by a name. It is such a kind of decorator pattern applied to HL Kripke structure.
 */
typedef struct HL_MODEL_s {
	/** the kripke structure of the model */
	KRIPKE_STRUCT_t* kripkeStruct;
	/** the table of propositional symbols */
	SYMBOL_TABLE_t* propSymTable;
	/** the table of nominals */
	SYMBOL_TABLE_t* nominalTable;
	/** the table of modalities of the model */
	SYMBOL_TABLE_t* modalityTable;
	/** the table of worlds of the model */
	SYMBOL_TABLE_t* worldTable;
	/** the name of the model */
	const char* modelName;
}  ///
HL_MODEL_t;

/**
 * Build a model parsing an XML file.
 * 
 * The file must be an instance of <code>'hl-model.dtd'</code>.
 * 
 * @param fileName the name of the XML file that contains the model
 * @return the model parsed from the file or <code>NULL</code> if an error is 
 * occurred
 */
HL_MODEL_t* HL_MODEL_new(const char *fileName);

/**
 * Print the Kripke struct.
 * 
 * @param os the output stream
 * @param model the model to print
 */
void HL_MODEL_printKripkeStruct(FILE* os, HL_MODEL_t* model);

/**
 * Pretty print a formula.
 * 
 * @param os the output stream
 * @param model the model to which the formula belongs
 * @param formula the formula to print
 * @wvarTable the symbol table of world variable in the formula
 */
void HL_MODEL_printFormula(FILE* os, HL_MODEL_t* model, HL_FORMULA_t* formula, SYMBOL_TABLE_t* wvarTable);

/**
 * Print the result of model checking procedure showing in which worlds of the model the formula holds.
 * 
 * @param os the output stream
 * @param model the model to which the formula belongs
 * @param formula the formula processed using the model checker
 */
void HL_MODEL_printMCResults(FILE* os, HL_MODEL_t* model, HL_FORMULA_t* formula);

/**
 * Delete a model deallocating memory.
 * 
 * @param model the model to destroy
 */
void HL_MODEL_delete(HL_MODEL_t* model);

#endif
