/*
 * Hybrid Logic Model Checker 0.9 - HL Formula data structures and Model Checking algorithms
 * 
 * Copyright (C) 2005 Luigi Dragone
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; version 2.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *   
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Luigi Dragone <luigi@luigidragone.com>
 */


#include <stdlib.h>
#include <string.h>
/** To import time-related functions and data types */
#include <time.h>

#include "hl.h"
#include "linked-list.h"

#define MAX(a, b) a > b ? a : b
#define HL_BOOL_PRINT(bool) bool ? "TRUE" : "FALSE"

static clock_t startTime;
static clock_t endTime;

static unsigned int mcFullLevelCount;
static unsigned int mcFullMaxLevelCount;
static unsigned int mcLiteInvocationCount;
static unsigned int mcFullInvocationCount;
static unsigned int mcDiaInvocationCount;
static unsigned int mcInvDiaInvocationCount;
static unsigned int mcExistsInvocationCount;
static unsigned int mcAtNomInvocationCount;
static unsigned int mcAtWVarInvocationCount;
static unsigned int checkBindInvocationCount;

typedef struct SYMBOL_TABLE_ENTRY_s {
	unsigned int identifier;
	const char* label;	
} SYMBOL_TABLE_ENTRY_t;

SYMBOL_TABLE_t* SYMBOL_TABLE_new() {
	SYMBOL_TABLE_t* t = malloc(sizeof(SYMBOL_TABLE_t));
	t->labelCount = 0;
	t->hashTable = pblHtCreate();
	return t;
}

void SYMBOL_TABLE_delete(SYMBOL_TABLE_t* t) {
	SYMBOL_TABLE_ENTRY_t* data = NULL;
	if(t != NULL) {
		for(data = pblHtFirst(t->hashTable); data != NULL; data = pblHtNext(t->hashTable)) {
			free((void*)data->label);
			free(data);
			pblHtRemove(t->hashTable, 0, 0);
		}
		pblHtDelete(t->hashTable);
		free(t);
	}
}

unsigned int SYMBOL_TABLE_addLabel(SYMBOL_TABLE_t* t, const char* label) {
	unsigned int labelId = HL_INVALID_LABEL;
	if(t != NULL && label != NULL) {
		size_t labelLen = strlen(label) + 1;
		SYMBOL_TABLE_ENTRY_t* entry = pblHtLookup(t->hashTable, (void*)label, labelLen);
		if(entry == NULL) {
			entry = malloc(sizeof(SYMBOL_TABLE_ENTRY_t));
			labelId = entry->identifier = t->labelCount;
			entry->label = strdup(label);
			pblHtInsert(t->hashTable, (void*)entry->label, labelLen, entry);
			t->labelCount++;
		} else
			labelId = entry->identifier;
	}
	return labelId;
}

size_t SYMBOL_TABLE_getLabelCount(SYMBOL_TABLE_t* t) {
	if(t != NULL)
		return t->labelCount;
	else
		return 0;
}

SYMBOL_ARRAY_t* SYMBOL_ARRAY_new(SYMBOL_TABLE_t* t) {
	SYMBOL_ARRAY_t* r = NULL;
	if(t != NULL) {
		r = malloc(sizeof(SYMBOL_ARRAY_t));
		r->entries = malloc(sizeof(const char*) * t->labelCount);
		r->length = t->labelCount;
		SYMBOL_TABLE_ENTRY_t* data = NULL;
		for(data = pblHtFirst(t->hashTable); data != NULL; data = pblHtNext(t->hashTable)) 
			r->entries[data->identifier] = strdup(data->label);
	}
	return r;
}

unsigned int SYMBOL_TABLE_getLabelIdentifier(SYMBOL_TABLE_t* t, const char* label) {
	unsigned int labelId = HL_INVALID_LABEL;
	if(t != NULL && label != NULL) {
		size_t labelLen = strlen(label) + 1;
		SYMBOL_TABLE_ENTRY_t* entry = pblHtLookup(t->hashTable, (void*)label, labelLen);
		if(entry != NULL) 
			labelId = entry->identifier;
	}
	return labelId;
}

const char* SYMBOL_ARRAY_getEntry(SYMBOL_ARRAY_t* a, unsigned int index) {
	if(a != NULL && index < a->length)
		return a->entries[index];
	else
		return NULL;
}

size_t SYMBOL_ARRAY_getLength(SYMBOL_ARRAY_t* a) {
	if(a != NULL)
		return a->length;
	else
		return 0;
}

void SYMBOL_ARRAY_delete(SYMBOL_ARRAY_t* a) {
	if(a != NULL) {
		int i;
		for(i = 0; i < a->length; i++)
			free((void*)a->entries[i]);
		free(a);	
	}
}

KRIPKE_STRUCT_t* KRIPKE_STRUCT_new(unsigned int worldNum, unsigned int modNum, unsigned int propSymNum, unsigned int nominalNum) {
	unsigned int i, j, k;
	KRIPKE_STRUCT_t* ks = malloc(sizeof(KRIPKE_STRUCT_t));
	ks->worldNum = worldNum;
	ks->modNum = modNum;
	ks->propSymNum = propSymNum;
	ks->nominalNum = nominalNum;
	if(nominalNum == 0)
		ks->nominalAssigns = NULL;
	else {
		ks->nominalAssigns = malloc(sizeof(unsigned int) * nominalNum);
		for(i = 0; i < nominalNum; i++)
			ks->nominalAssigns[i] = HL_UNASSIGNED;
	}
	ks->propSymAssigns = malloc(sizeof(HL_BOOLEAN*) * propSymNum);
	for(i = 0; i < propSymNum; i++) {
		ks->propSymAssigns[i] = malloc(sizeof(HL_BOOLEAN) * worldNum);
		for(j = 0; j < worldNum; j++)
			ks->propSymAssigns[i][j] = HL_FALSE;
	}
	ks->accessibilityRels = malloc(sizeof(HL_BOOLEAN**) * modNum);
	for(i = 0; i < modNum; i++) {
		ks->accessibilityRels[i] = malloc(sizeof(HL_BOOLEAN*) * worldNum);
		for(j = 0; j < worldNum; j++) {
			ks->accessibilityRels[i][j] = malloc(sizeof(HL_BOOLEAN) * worldNum);
			for(k = 0; k < worldNum; k++)
				ks->accessibilityRels[i][j][k] = HL_FALSE;
		}
	}
	return ks;
}

void KRIPKE_STRUCT_delete(KRIPKE_STRUCT_t* ks) {
	unsigned int i, j;
	if(ks->nominalNum != 0)
		free(ks->nominalAssigns);
	for(i = 0; i < ks->propSymNum; i++) {
		free(ks->propSymAssigns[i]);
	}
	free(ks->propSymAssigns);
	for(i = 0; i < ks->modNum; i++) {
		for(j = 0; j < ks->worldNum; j++)
			free(ks->accessibilityRels[i][j]);
		free(ks->accessibilityRels[i]);
	}
	free(ks->accessibilityRels);
	free(ks);
}

HL_BOOLEAN KRIPKE_STRUCT_isValidModalityId(KRIPKE_STRUCT_t* ks, unsigned int modalityId) {
	return ks->modNum > modalityId;	
}

HL_BOOLEAN KRIPKE_STRUCT_isValidWorldId(KRIPKE_STRUCT_t* ks, unsigned int worldId) {
	return ks->worldNum > worldId;	
}

HL_BOOLEAN KRIPKE_STRUCT_isValidPropSymId(KRIPKE_STRUCT_t* ks, unsigned int propSymId) {
	return ks->propSymNum > propSymId;	
}

HL_BOOLEAN KRIPKE_STRUCT_isValidNominalId(KRIPKE_STRUCT_t* ks, unsigned int nominalId) {
	return ks->nominalNum > nominalId;	
}

HL_BOOLEAN KRIPKE_STRUCT_isAccessible(KRIPKE_STRUCT_t* ks, unsigned int modalityId, unsigned int fromWorldId, unsigned int toWorldId) {
#ifdef DEBUG
	fprintf(stderr, "Checking accessibility from %d to %d on modality %d (%s)\n", fromWorldId, toWorldId, modalityId, HL_BOOL_PRINT(ks->accessibilityRels[modalityId][fromWorldId][toWorldId]));
#endif
	return ks->accessibilityRels[modalityId][fromWorldId][toWorldId];
}

void KRIPKE_STRUCT_setAccessibile(KRIPKE_STRUCT_t* ks, unsigned int modalityId, unsigned int fromWorldId, unsigned int toWorldId, HL_BOOLEAN accessibility) {
	ks->accessibilityRels[modalityId][fromWorldId][toWorldId] = accessibility;
}

HL_BOOLEAN KRIPKE_STRUCT_isPropSymTrue(KRIPKE_STRUCT_t* ks, unsigned int propSymId, unsigned int worldId) {
#ifdef DEBUG
	fprintf(stderr, "Checking truth assignment of propositional symbol %d at world %d (%s)\n", propSymId, worldId, HL_BOOL_PRINT(ks->propSymAssigns[propSymId][worldId]));
#endif
	return ks->propSymAssigns[propSymId][worldId];
}

void KRIPKE_STRUCT_setPropSym(KRIPKE_STRUCT_t* ks, unsigned int propSymId, unsigned int worldId, HL_BOOLEAN truthAssign) {
	ks->propSymAssigns[propSymId][worldId] = truthAssign;
}

unsigned int KRIPKE_STRUCT_getNominalAssign(KRIPKE_STRUCT_t* ks, unsigned int nominalId) {
#ifdef DEBUG
	fprintf(stderr, "Checking assignment of nominal %d (%d)\n", nominalId, ks->nominalAssigns[nominalId]);
#endif
	return ks->nominalAssigns[nominalId];
}

void KRIPKE_STRUCT_setNominalAssign(KRIPKE_STRUCT_t* ks, unsigned int nominalId, unsigned int worldId) {
	ks->nominalAssigns[nominalId] = worldId;
}

WVAR_ASSIGNMENT_t* WVAR_ASSIGNMENT_new(unsigned int wvarNum) {
	unsigned int i;
	WVAR_ASSIGNMENT_t* g = malloc(sizeof(WVAR_ASSIGNMENT_t));
	g->wvarNum = wvarNum;
	g->wvarAssigns = malloc(sizeof(unsigned int) * wvarNum);
	for(i = 0; i < wvarNum; i++)
		g->wvarAssigns[i] = HL_UNASSIGNED;
	return g;
}

void WVAR_ASSIGNMENT_delete(WVAR_ASSIGNMENT_t* g) {
	free(g->wvarAssigns);
	free(g);	
}

WVAR_ASSIGNMENT_t* WVAR_ASSIGNMENT_clone(WVAR_ASSIGNMENT_t* g) {
	size_t size;
	WVAR_ASSIGNMENT_t* g1 = malloc(sizeof(WVAR_ASSIGNMENT_t));
	g1->wvarNum = g->wvarNum;
	size = sizeof(unsigned int) * g->wvarNum;
	g1->wvarAssigns = malloc(size);
	memcpy(g1->wvarAssigns, g->wvarAssigns, size); 
	return g1;
}

unsigned int WVAR_ASSIGNMENT_getWVarAssign(WVAR_ASSIGNMENT_t* g, unsigned int wvarId) {
#ifdef DEBUG
	fprintf(stderr, "Getting assignment of wvar %d (%d)\n", wvarId, g->wvarAssigns[wvarId]);
#endif
	return g->wvarAssigns[wvarId];
}

void WVAR_ASSIGNMENT_setWVarAssign(WVAR_ASSIGNMENT_t* g, unsigned int wvarId, unsigned int worldId) {
#ifdef DEBUG
	fprintf(stderr, "Assign wvar %d to world %d\n", wvarId, worldId);
#endif
	g->wvarAssigns[wvarId] = worldId;
}

/***
 * Abstract type base constructor
 */
static HL_FORMULA_t* HL_FORMULA_new(KRIPKE_STRUCT_t* model) {
	unsigned int i;
	HL_FORMULA_t* f = malloc(sizeof(HL_FORMULA_t));
	f->parent = NULL;
	f->model = model;
	if(model == NULL)
		f->worldEvals = NULL;
	else {
		f->worldEvals = malloc(sizeof(HL_BOOLEAN) * model->worldNum);
		for(i = 0; i < model->worldNum; i++)
			f->worldEvals[i] = HL_FALSE;
	}
	return f;
}


HL_FORMULA_t* HL_FORMULA_newTrueConst(KRIPKE_STRUCT_t* model) {
	HL_FORMULA_t* f = NULL;
	f = HL_FORMULA_new(model);
	f->kind = trueK;
	f->size = 0;
	return f;	
}

HL_FORMULA_t* HL_FORMULA_newFalseConst(KRIPKE_STRUCT_t* model) {
	HL_FORMULA_t* f = NULL;
	f = HL_FORMULA_new(model);
	f->kind = falseK;
	f->size = 0;
	return f;	
}

HL_FORMULA_t* HL_FORMULA_newPropSym(KRIPKE_STRUCT_t* model, unsigned int propSymId) {
	HL_FORMULA_t* f = NULL;
	if(model == NULL || KRIPKE_STRUCT_isValidPropSymId(model, propSymId)) {
		f = HL_FORMULA_new(model);
		f->kind = propK;
		f->size = 1;
		f->val.identifier = propSymId;
	}
	return f;
}

HL_FORMULA_t* HL_FORMULA_newNominal(KRIPKE_STRUCT_t* model, unsigned int nominalId) {
	HL_FORMULA_t* f = NULL;
	if(model == NULL || KRIPKE_STRUCT_isValidNominalId(model, nominalId)) {
		f = HL_FORMULA_new(model);
		f->kind = nomK;
		f->size = 1;		
		f->val.identifier = nominalId;
	}
	return f;
}

HL_FORMULA_t* HL_FORMULA_newWVar(KRIPKE_STRUCT_t* model, unsigned int wvarId) {
	HL_FORMULA_t* f = HL_FORMULA_new(model);
	f->kind = wvarK;
	f->size = 1;
	f->val.identifier = wvarId;
	return f;
}

HL_FORMULA_t* HL_FORMULA_newNeg(HL_FORMULA_t* exp) {
	HL_FORMULA_t* f = NULL;
	if(exp != NULL && exp->parent == NULL) {
		f = HL_FORMULA_new(exp->model);
		exp->parent = f;
		f->kind = negK;
		f->size = exp->size + 1;
		f->val.negE.exp = exp;
	}
	return f;
}

HL_FORMULA_t* HL_FORMULA_newExists(HL_FORMULA_t* exp) {
	HL_FORMULA_t* f = NULL;
	if(exp != NULL && exp->parent == NULL) {
		f = HL_FORMULA_new(exp->model);
		exp->parent = f;
		f->kind = existsK;
		f->size = exp->size + 1;		
		f->val.existsE.exp = exp;
	}
	return f;
}

HL_FORMULA_t* HL_FORMULA_newAnd(HL_FORMULA_t* leftExp, HL_FORMULA_t* rightExp) {
	HL_FORMULA_t* f = NULL;
	if(leftExp != NULL && rightExp != NULL && leftExp->model == rightExp->model && leftExp->parent == NULL && rightExp->parent == NULL) {
		f = HL_FORMULA_new(leftExp->model);
		leftExp->parent = rightExp->parent = f;
		f->kind = andK;
		f->size = leftExp->size + rightExp->size + 1;
		f->val.andE.leftExp = leftExp;
		f->val.andE.rightExp = rightExp;
	}
	return f;
}

HL_FORMULA_t* HL_FORMULA_newDia(unsigned int modalityId, HL_FORMULA_t* exp) {
	HL_FORMULA_t* f = NULL;
	if(exp != NULL && exp->parent == NULL && (exp->model == NULL || KRIPKE_STRUCT_isValidModalityId(exp->model, modalityId) )) {
		f = HL_FORMULA_new(exp->model);
		exp->parent = f;
		f->kind = diaK;
		f->size = exp->size + 2;
		f->val.diaE.modalityId = modalityId;
		f->val.diaE.exp = exp;
	}
	return f;
}

HL_FORMULA_t* HL_FORMULA_newInvDia(unsigned int modalityId, HL_FORMULA_t* exp) {
	HL_FORMULA_t* f = NULL;
	if(exp != NULL && exp->parent == NULL && (exp->model == NULL || KRIPKE_STRUCT_isValidModalityId(exp->model, modalityId) )) {
		f = HL_FORMULA_new(exp->model);
		exp->parent = f;
		f->kind = invDiaK;
		f->size = exp->size + 2;
		f->val.invDiaE.modalityId = modalityId;
		f->val.invDiaE.exp = exp;
	}
	return f;
}

HL_FORMULA_t* HL_FORMULA_newAtNom(unsigned int nominalId, HL_FORMULA_t* exp) {
	HL_FORMULA_t* f = NULL;
	if(exp != NULL && exp->parent == NULL && (exp->model == NULL || KRIPKE_STRUCT_isValidNominalId(exp->model, nominalId) )) {
		f = HL_FORMULA_new(exp->model);
		exp->parent = f;
		f->kind = atNomK;
		f->size = exp->size + 2;
		f->val.atNomE.nominalId = nominalId;
		f->val.atNomE.exp = exp;
	}
	return f;
}

HL_FORMULA_t* HL_FORMULA_newAtWVar(unsigned int wvarId, HL_FORMULA_t* exp) {
	HL_FORMULA_t* f = NULL;
	if(exp != NULL && exp->parent == NULL) {
		f = HL_FORMULA_new(exp->model);
		exp->parent = f;
		f->kind = atWVarK;
		f->size = exp->size + 2;		
		f->val.atWVarE.wvarId = wvarId;
		f->val.atWVarE.exp = exp;
	}
	return f;
}

HL_FORMULA_t* HL_FORMULA_newBindWVar(unsigned int wvarId, HL_FORMULA_t* exp) {
	HL_FORMULA_t* f = NULL;
	if(exp != NULL && exp->parent == NULL) {
		f = HL_FORMULA_new(exp->model);
		exp->parent = f;
		f->kind = bindWVarK;
		f->size = exp->size + 2;		
		f->val.bindWVarE.wvarId = wvarId;
		f->val.bindWVarE.exp = exp;
	}
	return f;
}

void HL_FORMULA_delete(HL_FORMULA_t* exp) {
	if(exp != NULL) {
		switch(exp->kind) {
		case negK:
			HL_FORMULA_delete(exp->val.negE.exp);
			break;
		case andK:
			HL_FORMULA_delete(exp->val.andE.rightExp);
			HL_FORMULA_delete(exp->val.andE.leftExp);
			break;
		case atNomK:
			HL_FORMULA_delete(exp->val.atNomE.exp);
			break;
		case atWVarK:
			HL_FORMULA_delete(exp->val.atWVarE.exp);
			break;
		case bindWVarK:
			HL_FORMULA_delete(exp->val.bindWVarE.exp);
			break;
		case diaK:
			HL_FORMULA_delete(exp->val.diaE.exp);
			break;
		case invDiaK:
			HL_FORMULA_delete(exp->val.invDiaE.exp);
			break;
		case existsK:
			HL_FORMULA_delete(exp->val.existsE.exp);
			break;
		default:
			break;
		}		
		if(exp->model != NULL && exp->model->worldNum != 0)
			free(exp->worldEvals);
		free(exp);
	}
}

/*
 * for beta  sub(alpha) do
 * 	for w  W do
 *   L(beta, w) := FALSE
 *  end for
 * end for
 */
static void HL_FORMULA_resetWorldEvals(HL_FORMULA_t* exp, HL_BOOLEAN recursively) {
	int i;
	if(exp != NULL && exp->worldEvals != NULL && exp->model != NULL) {
		for(i = 0; i < exp->model->worldNum; i++)
			exp->worldEvals[i] = HL_FALSE;	
		if(recursively) 
			switch(exp->kind) {
			case negK:
				HL_FORMULA_resetWorldEvals(exp->val.negE.exp, recursively);
				break;
			case andK:
				HL_FORMULA_resetWorldEvals(exp->val.andE.rightExp, recursively);
				HL_FORMULA_resetWorldEvals(exp->val.andE.leftExp, recursively);
				break;
			case atNomK:
				HL_FORMULA_resetWorldEvals(exp->val.atNomE.exp, recursively);
				break;
			case atWVarK:
				HL_FORMULA_resetWorldEvals(exp->val.atWVarE.exp, recursively);
				break;
			case bindWVarK:
				HL_FORMULA_resetWorldEvals(exp->val.bindWVarE.exp, recursively);
				break;
			case diaK:
				HL_FORMULA_resetWorldEvals(exp->val.diaE.exp, recursively);
				break;
			case invDiaK:
				HL_FORMULA_resetWorldEvals(exp->val.invDiaE.exp, recursively);
				break;
			case existsK:
				HL_FORMULA_resetWorldEvals(exp->val.existsE.exp, recursively);
				break;
			default:
				break;
			}		
	}
}

/**
 * Count the number of different variable symbols occurring in a formula.
 * 
 * @param exp the formula to process
 * @return the number of different variable symbols
 */
static unsigned int HL_FORMULA_countWVar(HL_FORMULA_t* exp) {
	int count = 0;
	if(exp != NULL) {
		switch(exp->kind) {
		case wvarK:
			count = exp->val.identifier + 1;
			break;
		case negK:
			count = HL_FORMULA_countWVar(exp->val.negE.exp);
			break;
		case andK:
			count = MAX(HL_FORMULA_countWVar(exp->val.andE.rightExp), HL_FORMULA_countWVar(exp->val.andE.leftExp));
			break;
		case atNomK:
			count = HL_FORMULA_countWVar(exp->val.atNomE.exp);
			break;
		case atWVarK:
			count = MAX(exp->val.atWVarE.wvarId + 1, HL_FORMULA_countWVar(exp->val.atWVarE.exp));
			break;
		case bindWVarK:
			count = MAX(exp->val.bindWVarE.wvarId + 1, HL_FORMULA_countWVar(exp->val.bindWVarE.exp));
			break;
		case diaK:
			count = HL_FORMULA_countWVar(exp->val.diaE.exp);
			break;
		case invDiaK:
			count = HL_FORMULA_countWVar(exp->val.invDiaE.exp);
			break;
		case existsK:
			count = HL_FORMULA_countWVar(exp->val.existsE.exp);
			break;
		default:
			break;
		}		
	}
	return count;
}	

/*
 * clear L(alpha, w)
 */ 
static HL_BOOLEAN HL_FORMULA_clear(HL_FORMULA_t* exp, unsigned int wvarId) {
	HL_BOOLEAN wvarFound = HL_FALSE;
#ifdef DEBUG
	fprintf(stderr, "Clearing wvar binding\n");
#endif
	if(exp != NULL) {
		switch(exp->kind) {
			case wvarK:
				wvarFound = exp->val.identifier == wvarId;
				break;
			case andK:
				wvarFound = HL_FORMULA_clear(exp->val.andE.leftExp, wvarId) || HL_FORMULA_clear(exp->val.andE.rightExp, wvarId);
				break;
			case negK:
				wvarFound = HL_FORMULA_clear(exp->val.negE.exp, wvarId);
				break;
			case diaK:
				wvarFound = HL_FORMULA_clear(exp->val.diaE.exp, wvarId);
				break;
			case invDiaK:
				wvarFound = HL_FORMULA_clear(exp->val.invDiaE.exp, wvarId);
				break;
			case existsK:
				wvarFound = HL_FORMULA_clear(exp->val.existsE.exp, wvarId);
				break;
			case atNomK:
				wvarFound = HL_FORMULA_clear(exp->val.atNomE.exp, wvarId);
				break;
			case atWVarK:
				wvarFound = exp->val.atWVarE.wvarId == wvarId || HL_FORMULA_clear(exp->val.atWVarE.exp, wvarId);
				break;
			case bindWVarK:
				wvarFound = exp->val.bindWVarE.wvarId == wvarId || HL_FORMULA_clear(exp->val.bindWVarE.exp, wvarId);
				break;
			default:	
				break;	
		}
		if(wvarFound)
			HL_FORMULA_resetWorldEvals(exp,HL_FALSE);	
	}
	return wvarFound;	
}

/*
 * MC<> (alpha)
 */
static void HL_FORMULA_mcDia(HL_FORMULA_t* exp) {
	unsigned int worldId_1, worldId_2;
	unsigned int modalityId = exp->parent->val.diaE.modalityId;
	mcDiaInvocationCount++;
#ifdef DEBUG
	fprintf(stderr, "Evaluating MC<>\n");
#endif
	// for w  L(alpha) do
	for(worldId_1 = 0; worldId_1 < exp->model->worldNum; worldId_1++) 
		if(exp->worldEvals[worldId_1]) 
			// for v  R-1(w) do
			for(worldId_2 = 0; worldId_2 < exp->model->worldNum; worldId_2++)
				if(KRIPKE_STRUCT_isAccessible(exp->model, modalityId, worldId_2, worldId_1))	
					// L(<>alpha, v) := TRUE
					exp->parent->worldEvals[worldId_2] = HL_TRUE;		
			// end for
	// end for
}

/*
 * MC<>- (alpha)
 */
static void HL_FORMULA_mcInvDia(HL_FORMULA_t* exp) {
	unsigned int worldId_1, worldId_2;
	unsigned int modalityId = exp->parent->val.diaE.modalityId;
	mcInvDiaInvocationCount++;
#ifdef DEBUG
	fprintf(stderr, "Evaluating MC<>-\n");
#endif
	// for w  L(alpha) do
	for(worldId_1 = 0; worldId_1 < exp->model->worldNum; worldId_1++) 
		if(exp->worldEvals[worldId_1]) 
			// for v  R(w) do
			for(worldId_2 = 0; worldId_2 < exp->model->worldNum; worldId_2++) 
				if(KRIPKE_STRUCT_isAccessible(exp->model, modalityId, worldId_1, worldId_2))	
					// L(<>-alpha, v) := TRUE
					exp->parent->worldEvals[worldId_2] = HL_TRUE;		
			// end for
	// end for
}

/*
 * MC_E(alpha)
 */
static void HL_FORMULA_mcExists(HL_FORMULA_t* exp) {
	unsigned int worldId_1, worldId_2;
	mcExistsInvocationCount++;
#ifdef DEBUG
	fprintf(stderr, "Evaluating MCE\n");
#endif
	// if L(alpha) =/= 0 then
	for(worldId_1 = 0; worldId_1 < exp->model->worldNum; worldId_1++) 
		if(exp->worldEvals[worldId_1]) { 
			// for w  W do
			for(worldId_2 = 0; worldId_2 < exp->model->worldNum; worldId_2++) 
				// L(Ealpha, w) := TRUE
				exp->parent->worldEvals[worldId_2] = HL_TRUE;
			// end for
			return;
		}
	// end if
}

/*
 * MC_@(i, alpha)
 */
void HL_FORMULA_mcAtNom(HL_FORMULA_t* exp) {
	// i
	unsigned int nominalId = exp->parent->val.atNomE.nominalId;
	// let v = V(i)
	unsigned int worldId_1 = KRIPKE_STRUCT_getNominalAssign(exp->model, nominalId);
	unsigned int worldId_2;
	
	mcAtNomInvocationCount++;
#ifdef DEBUG
	fprintf(stderr, "Evaluating MC@nom\n");
#endif
	// if L(alpha, v) then
	if(exp->worldEvals[worldId_1])
		// for w  W do
		for(worldId_2 = 0; worldId_2 < exp->model->worldNum; worldId_2++) 
			// L(@i alpha, w) := TRUE
			exp->parent->worldEvals[worldId_2] = HL_TRUE;
		// end for
	// end if
}

/*
 * MC_@(x, alpha)
 */
static void HL_FORMULA_mcAtWVar(HL_FORMULA_t* exp, WVAR_ASSIGNMENT_t* g) {
	// x
	unsigned int wvarId = exp->parent->val.atWVarE.wvarId;
	// let v = g(x)
	unsigned int worldId_1 = WVAR_ASSIGNMENT_getWVarAssign(g, wvarId);
	unsigned int worldId_2;
	
	mcAtWVarInvocationCount++;
#ifdef DEBUG
	fprintf(stderr, "Evaluating MC@wvar\n");
#endif
	// if L(alpha, v) then
	if(exp->worldEvals[worldId_1])
		// for w  W do
		for(worldId_2 = 0; worldId_2 < exp->model->worldNum; worldId_2++) 
			// L(@x alpha, w) := TRUE
			exp->parent->worldEvals[worldId_2] = HL_TRUE;
		// end for
	// end if
}

/* Forward reference */
static void HL_FORMULA_mcFull(HL_FORMULA_t* exp, WVAR_ASSIGNMENT_t* g);

/*
 * Check^(g, x, alpha)
 */
static void HL_FORMULA_checkBind(HL_FORMULA_t* exp, WVAR_ASSIGNMENT_t* g) {
	// x
	unsigned wvarId = exp->parent->val.bindWVarE.wvarId;
	unsigned int worldId;
	
	checkBindInvocationCount++;
#ifdef DEBUG
	fprintf(stderr, "Checking binding of wvar %d\n", wvarId);
#endif
	WVAR_ASSIGNMENT_t* g1 = WVAR_ASSIGNMENT_clone(g);
	// for w  W do
	for(worldId = 0; worldId < exp->model->worldNum; worldId++) {
		// g(x) := w
		WVAR_ASSIGNMENT_setWVarAssign(g1, wvarId, worldId);
		// MCFull(M, g, alpha)
		HL_FORMULA_mcFull(exp, g1);
#ifdef DEBUG
	fprintf(stderr, "Binding to world %d is %s\n", worldId, HL_BOOL_PRINT(exp->worldEvals[worldId]));
#endif		
		// if L(alpha, w) then
		if(exp->worldEvals[worldId]) {
			// Clear(L, x)
			HL_FORMULA_clear(exp, wvarId);
			// L(^x.alpha, w) := TRUE
			exp->parent->worldEvals[worldId] = HL_TRUE;			
		// else
		} else
			// Clear(L, x)
			HL_FORMULA_clear(exp, wvarId);
		// end if
	}
	// end for
	WVAR_ASSIGNMENT_delete(g1);
}

/*
 * MCFull (g, alpha)
 */
static void HL_FORMULA_mcFull(HL_FORMULA_t* exp, WVAR_ASSIGNMENT_t* g) {
	unsigned int worldId;
	mcFullInvocationCount++;
	mcFullLevelCount++;
#ifdef DEBUG
	fprintf(stderr, "Evaluatating MCFull at level %d\n", mcFullLevelCount);
#endif
	if(mcFullLevelCount > mcFullMaxLevelCount)
		mcFullMaxLevelCount = mcFullLevelCount;
	HL_FORMULA_resetWorldEvals(exp, HL_TRUE);
	// case
	switch(exp->kind) {
		// * alpha = T
		case trueK:
			// for w  W do
			for(worldId = 0; worldId < exp->model->worldNum; worldId++)
				// L(alpha, w) := TRUE
				exp->worldEvals[worldId] = HL_TRUE;
			// end for
			break;
		// * alpha = F
		case falseK:
			// for w  W do
			for(worldId = 0; worldId < exp->model->worldNum; worldId++)
				// L(alpha, w) := FALSE
				exp->worldEvals[worldId] = HL_FALSE;
			// end for
			break;
		// * alpha  PROP
		case propK:
			// for w  W do
			for(worldId = 0; worldId < exp->model->worldNum; worldId++)
				// if w  V(alpha) then L(alpha, w) := TRUE end if
				exp->worldEvals[worldId] = KRIPKE_STRUCT_isPropSymTrue(exp->model, exp->val.identifier, worldId);
			// end for
			break;
		// * alpha  NOM
		case nomK:
			// for w  W do 
			worldId = KRIPKE_STRUCT_getNominalAssign(exp->model, exp->val.identifier);
			// if w  V(alpha) then L(alpha, w) := TRUE end if 
			exp->worldEvals[worldId] = HL_TRUE;
			// end if
			// end for
			break;
		// * alpha  WVAR
		case wvarK:
			// L(alpha, g(alpha)) := TRUE
			worldId = WVAR_ASSIGNMENT_getWVarAssign(g, exp->val.identifier);
			exp->worldEvals[worldId] = HL_TRUE;
			break;
		// * alpha = alpha_1 and alpha_2
		case andK:
			// MCFull(g, alpha_1)
			HL_FORMULA_mcFull(exp->val.andE.leftExp, g);
			// MCFull(g, alpha_2)
			HL_FORMULA_mcFull(exp->val.andE.rightExp, g);
			// for w  W do
			for(worldId = 0; worldId < exp->model->worldNum; worldId++) 
				// if L(alpha_1, w) and L(alpha_2, w) then
				if(exp->val.andE.leftExp->worldEvals[worldId] && exp->val.andE.rightExp->worldEvals[worldId])
					// L(alpha, w) := TRUE
					exp->worldEvals[worldId] = HL_TRUE;
				// end if
			// end for
			break;
		// * alpha = not alpha_1
		case negK:
			// MCFull(g, alpha_1)
			HL_FORMULA_mcFull(exp->val.negE.exp, g);
			// for w  W do
			for(worldId = 0; worldId < exp->model->worldNum; worldId++) 
				// if not L(alpha_1, w) then
				if(!exp->val.negE.exp->worldEvals[worldId])
					// L(alpha, w) := TRUE
					exp->worldEvals[worldId] = HL_TRUE;
				// end if
			// end for
			break;
		// * alpha = <pi> alpha_1
		case diaK:
			// MCFull(g, alpha_1)
			HL_FORMULA_mcFull(exp->val.diaE.exp, g);
			// MC<>(alpha_1)
			HL_FORMULA_mcDia(exp->val.diaE.exp);
			break;
		// * alpha = <pi>- alpha_1
		case invDiaK:
			// MCFull(g, alpha_1)
			HL_FORMULA_mcFull(exp->val.invDiaE.exp, g);
			// MC<>-(alpha_1)
			HL_FORMULA_mcInvDia(exp->val.invDiaE.exp);
			break;
		// * alpha = E alpha_1
		case existsK:
			// MCFull(g, alpha_1)
			HL_FORMULA_mcFull(exp->val.existsE.exp, g);
			// MC_E(alpha_1)
			HL_FORMULA_mcExists(exp->val.existsE.exp);
			break;
		// * alpha = @i alpha_1
		case atNomK:
			// MCFull(g, alpha_1)
			HL_FORMULA_mcFull(exp->val.atNomE.exp, g);
			// MC_@(i, alpha)
			HL_FORMULA_mcAtNom(exp->val.atNomE.exp);
			break;
		// * alpha = @x alpha_1
		case atWVarK:
			// MCFull(g, alpha_1)
			HL_FORMULA_mcFull(exp->val.atWVarE.exp, g);
			// MC_@(x, alpha)
			HL_FORMULA_mcAtWVar(exp->val.atWVarE.exp, g);
			break;	
		// * alpha = ^x.alpha_1	
		case bindWVarK:
			// Check^(g, x, alpha_1)
			HL_FORMULA_checkBind(exp->val.bindWVarE.exp, g);
		default:	
			break;	
	}
	mcFullLevelCount--;
}
	

HL_BOOLEAN* HL_FORMULA_getWorldEvals(HL_FORMULA_t* exp) {
	if(exp != NULL)
		return exp->worldEvals;
	else
		return NULL;
}

void HL_FORMULA_prettyPrint(FILE* os, HL_FORMULA_t* exp, SYMBOL_ARRAY_t* propSymLabels, 
	SYMBOL_ARRAY_t* nominalLabels, SYMBOL_ARRAY_t* modalityLabels, SYMBOL_ARRAY_t* wvarLabels) {
	if(exp != NULL) {
		switch(exp->kind) {
		case trueK:
			fprintf(os, "T");
			break;
		case falseK:
			fprintf(os, "F");
			break;
		case propK:
			fprintf(os, "%s", propSymLabels->entries[exp->val.identifier]);
			break;
		case nomK:
			fprintf(os, "%s", nominalLabels->entries[exp->val.identifier]);
			break;
		case wvarK:
			fprintf(os, "%s", wvarLabels->entries[exp->val.identifier]);
			break;
		case negK:
			fprintf(os, "!(");
			HL_FORMULA_prettyPrint(os, exp->val.negE.exp, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
			fprintf(os, ")");			
			break;
		case andK:
			fprintf(os, "(");
			HL_FORMULA_prettyPrint(os, exp->val.andE.leftExp, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
			fprintf(os, " & ");
			HL_FORMULA_prettyPrint(os, exp->val.andE.rightExp, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
			fprintf(os, ")");			
			break;
		case atNomK:
			fprintf(os, "@ %s (", nominalLabels->entries[exp->val.atNomE.nominalId]);
			HL_FORMULA_prettyPrint(os, exp->val.atNomE.exp, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
			fprintf(os, ")");			
			break;
		case atWVarK:
			fprintf(os, "@ %s (", wvarLabels->entries[exp->val.atWVarE.wvarId]);
			HL_FORMULA_prettyPrint(os, exp->val.atWVarE.exp, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
			fprintf(os, ")");			
			break;
		case bindWVarK:
			fprintf(os, "B %s (", wvarLabels->entries[exp->val.bindWVarE.wvarId]);
			HL_FORMULA_prettyPrint(os, exp->val.bindWVarE.exp, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
			fprintf(os, ")");			
			break;
		case diaK:
			fprintf(os, "<%s>(", modalityLabels->entries[exp->val.diaE.modalityId]);
			HL_FORMULA_prettyPrint(os, exp->val.diaE.exp, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
			fprintf(os, ")");			
			break;
		case invDiaK:
			fprintf(os, "<%s>-(", modalityLabels->entries[exp->val.invDiaE.modalityId]);
			HL_FORMULA_prettyPrint(os, exp->val.invDiaE.exp, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
			fprintf(os, ")");			
			break;
		case existsK:
			fprintf(os, "E(");
			HL_FORMULA_prettyPrint(os, exp->val.existsE.exp, propSymLabels, nominalLabels, modalityLabels, wvarLabels);
			fprintf(os, ")");			
			break;
		}		
	}
}

/*
 *  sub(alpha)
 */
static LINKED_LIST_t* HL_FORMULA_subFormulaList(HL_FORMULA_t* exp, LINKED_LIST_t* tail) {
	LINKED_LIST_t* subFormulaList = NULL;
	if(exp != NULL) {
		switch(exp->kind) {
		case trueK:
		case falseK:
		case propK:
		case nomK:
		case wvarK:
			subFormulaList = LINKED_LIST_new(exp, tail);
			break;
		case negK:
			tail = HL_FORMULA_subFormulaList(exp->val.negE.exp, tail);
			subFormulaList = LINKED_LIST_new(exp, tail);
			break;
		case andK:
			tail = HL_FORMULA_subFormulaList(exp->val.andE.leftExp, tail);
			tail = HL_FORMULA_subFormulaList(exp->val.andE.rightExp, tail);
			subFormulaList = LINKED_LIST_new(exp, tail);
			break;
		case atNomK:
			tail = HL_FORMULA_subFormulaList(exp->val.atNomE.exp, tail);
			subFormulaList = LINKED_LIST_new(exp, tail);
			break;
		case atWVarK:
			tail = HL_FORMULA_subFormulaList(exp->val.atWVarE.exp, tail);
			subFormulaList = LINKED_LIST_new(exp, tail);
			break;
		case bindWVarK:
			tail = HL_FORMULA_subFormulaList(exp->val.bindWVarE.exp, tail);
			subFormulaList = LINKED_LIST_new(exp, tail);
			break;
		case diaK:
			tail = HL_FORMULA_subFormulaList(exp->val.diaE.exp, tail);
			subFormulaList = LINKED_LIST_new(exp, tail);
			break;
		case invDiaK:
			tail = HL_FORMULA_subFormulaList(exp->val.invDiaE.exp, tail);
			subFormulaList = LINKED_LIST_new(exp, tail);
			break;
		case existsK:
			tail = HL_FORMULA_subFormulaList(exp->val.existsE.exp, tail);
			subFormulaList = LINKED_LIST_new(exp, tail);
			break;
		}		
	}	
	return subFormulaList;
}

/*
 * MCLite (alpha)
 */
static void HL_FORMULA_mcLite(HL_FORMULA_t* alpha) {
	unsigned int worldId;
#ifdef DEBUG
	fprintf(stderr, "Evaluatating MCLite\n");
#endif
	unsigned int i;
	LINKED_LIST_t* listPtr;
	HL_FORMULA_t* beta;
	LINKED_LIST_t* subFormulas = HL_FORMULA_subFormulaList(alpha, NULL);
	size_t maxSize = alpha->size;
	HL_FORMULA_resetWorldEvals(alpha, HL_TRUE);
	// for i from 1 to |alpha| do
	for(i = 0; i <= maxSize; i++) {
		// for beta  sub(alpha) do
		for(listPtr = subFormulas; listPtr != NULL; listPtr = listPtr->next) {
			beta = listPtr->data;
			// if |beta| = i then
			if(beta->size == i) {
				mcLiteInvocationCount++;
				// case
				switch(beta->kind) {
					// * beta = T
					case trueK:
						// for w  W do
						for(worldId = 0; worldId < beta->model->worldNum; worldId++)
							// L(beta, w) := TRUE
							beta->worldEvals[worldId] = HL_TRUE;
						// end for
						break;
					// * beta = F
					case falseK:
						// for w  W do
						for(worldId = 0; worldId < beta->model->worldNum; worldId++)
							// L(beta, w) := FALSE
							beta->worldEvals[worldId] = HL_FALSE;
						// end for
						break;
					// * beta  PROP
					case propK:
						// for w  W do
						for(worldId = 0; worldId < beta->model->worldNum; worldId++)
							// if w  V(beta) then L(beta, w) := TRUE end if
							beta->worldEvals[worldId] = KRIPKE_STRUCT_isPropSymTrue(beta->model, beta->val.identifier, worldId);
						// end for
						break;
					// * beta  NOM
					case nomK:
						// for w  W do 
						worldId = KRIPKE_STRUCT_getNominalAssign(beta->model, beta->val.identifier);
						// if w  V(beta) then L(beta, w) := TRUE end if 
						beta->worldEvals[worldId] = HL_TRUE;
						// end if
						// end for
						break;
					// * beta = beta_1 and beta_2
					case andK:
						// for w  W do
						for(worldId = 0; worldId < beta->model->worldNum; worldId++) 
							// if L(beta_1, w) and L(beta_2, w) then
							if(beta->val.andE.leftExp->worldEvals[worldId] && beta->val.andE.rightExp->worldEvals[worldId])
								// L(beta, w) := TRUE
								beta->worldEvals[worldId] = HL_TRUE;
							// end if
						// end for
						break;
					// * beta = not beta_1
					case negK:
						// for w  W do
						for(worldId = 0; worldId < beta->model->worldNum; worldId++) 
							// if not L(beta_1, w) then
							if(!beta->val.negE.exp->worldEvals[worldId])
								// L(beta, w) := TRUE
								beta->worldEvals[worldId] = HL_TRUE;
							// end if
						// end for
						break;
					// * beta = <pi> beta_1
					case diaK:
						// MC<>(beta_1)
						HL_FORMULA_mcDia(beta->val.diaE.exp);
						break;
					// * beta = <pi>- beta_1
					case invDiaK:
						// MC<>-(beta_1)
						HL_FORMULA_mcInvDia(beta->val.invDiaE.exp);
						break;
					// * beta = E beta_1
					case existsK:
						// MC_E(beta_1)
						HL_FORMULA_mcExists(beta->val.existsE.exp);
						break;
					// * beta = @i beta_1
					case atNomK:
						// MC_@(i, beta)
						HL_FORMULA_mcAtNom(beta->val.atNomE.exp);
						break;
					default:
						fprintf(stderr, "Unsupported formula kind, please switch to MCFull!\n");	
						return;	
				}
			// end if
			} 
		// end for
		}
	// end for
	}
	LINKED_LIST_delete(subFormulas);
}

/***
 * Reset the invocation counters of MC algorithms
 */
static void HL_FORMULA_resetMCCounters() {
	mcLiteInvocationCount = mcFullInvocationCount = mcDiaInvocationCount = mcInvDiaInvocationCount = mcExistsInvocationCount = mcAtNomInvocationCount = mcAtWVarInvocationCount = checkBindInvocationCount = 
	mcFullLevelCount = mcFullMaxLevelCount = 0;	
}

void HL_FORMULA_printMCStats(FILE* os, HL_BOOLEAN useMCLite) {
	fprintf(os, "Evaluation stats:\n");
	fprintf(os, "  Running time\t\t%.4f (s)\n", (double)(endTime - startTime)/CLOCKS_PER_SEC);
	fprintf(os, "  Invocation counters:\n");
	if(useMCLite)
		fprintf(os, "    MCLite\t\t%d\n", mcLiteInvocationCount);
	else 
		fprintf(os, "    MCFull\t\t%d\n", mcFullInvocationCount);
	fprintf(os, "    MC<>\t\t%d\n", mcDiaInvocationCount);
	fprintf(os, "    MC<>-\t\t%d\n", mcInvDiaInvocationCount);
	fprintf(os, "    MCE\t\t\t%d\n", mcExistsInvocationCount);
	fprintf(os, "    MC@nom\t\t%d\n", mcAtNomInvocationCount);
	if(!useMCLite) {
		fprintf(os, "    MC@wvar\t\t%d\n", mcAtWVarInvocationCount);
		fprintf(os, "    CheckBind\t\t%d\n", checkBindInvocationCount);
		fprintf(os, "  Maximum nesting\t%d\n", mcFullMaxLevelCount);
	}
	
}

void HL_FORMULA_modelChecking(HL_FORMULA_t* exp, HL_BOOLEAN useMCLite) {
	HL_FORMULA_resetMCCounters();
	startTime = clock();
	if(useMCLite) {
#ifdef DEBUG
		fprintf(stderr, "MCLite procedure started\n");
#endif
		HL_FORMULA_mcLite(exp);
#ifdef DEBUG
		fprintf(stderr, "MCLite procedure completed\n");
#endif
	} else {
		int wvarNum = HL_FORMULA_countWVar(exp);
		WVAR_ASSIGNMENT_t* g = WVAR_ASSIGNMENT_new(wvarNum);
#ifdef DEBUG
		fprintf(stderr, "%d world variable assignment object allocated\n", wvarNum);
		fprintf(stderr, "MCFull procedure started\n");
#endif
		HL_FORMULA_mcFull(exp, g);
#ifdef DEBUG
		fprintf(stderr, "MCFull procedure completed\n");
#endif
		WVAR_ASSIGNMENT_delete(g);
	}
	endTime = clock();
}
