/*
 * Hybrid Logic Model Checker 0.9
 * 
 * Copyright (C) 2005 Luigi Dragone
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; version 2.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *   
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Luigi Dragone <luigi@luigidragone.com>
 */

/// To use I/O functions
#include <stdio.h>
/// To use error management and reporting functions
#include <errno.h>
/// To use command line option parser functions
#include <unistd.h>
/// To use HL functions and data structures
#include "hl.h"
/// To parse HL model XML files
#include "hl-model.h"

/// HL formula parser's line counter
int lineno;
/// HL formula build by the parser
HL_FORMULA_t* theFormula;
/// Kripke structure used by HL formula parser
KRIPKE_STRUCT_t* kripkeStruct;
/// Table of propositional symbols used by HL formula parser
SYMBOL_TABLE_t* propSymTable; 
/// Table of nominals used by HL formula parser
SYMBOL_TABLE_t* nominalTable; 
/// Table of modalities used by HL formula parser
SYMBOL_TABLE_t* modalityTable; 
/// Table of world variables used by HL formula parser
SYMBOL_TABLE_t* wvarTable;
/** HL formula parser.
 * 
 * It parses the standard input and returns parsed the formula in the global
 * variable <code>theFormula</code>. It the requires that the following global 
 * variables are properly initialized before its execution:
 * <ul>
 * 	<li><code>kripkeStruct</code></li>
 * 	<li><code>propSymTable</code></li>
 * 	<li><code>nominalTable</code></li>
 * 	<li><code>modalityTable</code></li>
 * 	<li><code>wvarTable</code></li>
 * </ul>
 * If an error occurs during the parsing the output variable is set to 
 * <code>NULL</code>.
 * The formula is expressed according to grammar in <code>hl-formula.bnf</code>
 */
void yyparse();

/**
 * Print general information about the program, command syntax and options
 */
static void printCommandLineReference() {
	printf("Hybrid Logic Model Checker 0.9\n");
	printf("Copyright (C) 2005 Luigi Dragone <luigi@luigidragone.com>\n");
	printf("\nThis program is free software; you can redistribute it and/or modify ");
	printf("it under the terms of the GNU General Public License as published by ");
	printf("the Free Software Foundation; version 2.\n");
	printf("\nUsage: hlmc [-h] | ([-l] [-s] <model_file>)\n");
	printf("\nOptions:\n");
	printf("\t-h            shows program usage and reference information\n");
	printf("\t-l            switchs to MCLite procedure (otherwise MCFull is used)\n");
	printf("\t-s            prints evaluation statistics\n");
	printf("\t<model_file>  is the name of the XML file, instance of 'hl-ks.dtd', \n");
	printf("\t              containing the Kripke structure to use as model\n");
	printf("\nReturn codes:\n");
	printf("\t0 if everything is OK\n");
	printf("\t1 if an error is occurred\n");
	printf("\nThis program reads a formula from the stdin and checks in which worlds\n");
	printf("of the given model the formula holds.\n");	
}

/**
 * Main function.
 * 
 * It performs the model checking of HL formula according to following procedure:
 * <ol>
 * <li>read the XML file containing the HL model</li>
 * <li>parse the HL formula from stdin</li>
 * <li>init data structures</li>
 * <li>run the MC procedure (Full or Lite)</li>
 * <li>print the results on stdout</li>
 * </ol>
 */
int main(int argc, char * const argv[]) {
	const char* modelFileName = NULL;
	HL_BOOLEAN useMCLite = HL_FALSE;
	HL_BOOLEAN printStats = HL_FALSE;
	int option;
	
	opterr = 0;
	while((option = getopt(argc, argv, "lsh")) != -1) {
		switch(option) {
		case 'l':
#ifdef DEBUG
			fprintf(stderr, "MCLite procedure enabled, only no world variables allowed!\n");
#endif
			useMCLite = HL_TRUE;
			break;
		case 's':
#ifdef DEBUG			
			fprintf(stderr, "Evaluation statistics enabled.\n");
#endif
			printStats = HL_TRUE;
			break;
		case 'h':
			printCommandLineReference();
			return 0;
		default:
			fprintf(stderr, "Invalid option!\n"); 
			printCommandLineReference();
			return 1;
		}	
	}

	if(argc == optind || argv[optind] == NULL) {
		fprintf(stderr, "Missing model file name!\n"); 
		printCommandLineReference();
		return 1;
	}
	modelFileName = argv[optind];
#ifdef DEBUG
	fprintf(stderr, "Reading the model from \'%s\'...\n", modelFileName);
#endif
	HL_MODEL_t* theModel = HL_MODEL_new(modelFileName);
	if(theModel == NULL) {
		fprintf(stderr, "Error processing the model!\n");
		return 1;	
	}
	
#ifdef DEBUG
	HL_MODEL_printKripkeStruct(stdout, theModel);
#endif

#ifdef DEBUG
	fprintf(stderr, "Parsing the formula from stdin...\n");
#endif
	wvarTable = SYMBOL_TABLE_new();
	propSymTable = theModel->propSymTable; 
	nominalTable = theModel->nominalTable; 
	modalityTable = theModel->modalityTable; 
	kripkeStruct = theModel->kripkeStruct;
	yyparse();
	if(theFormula == NULL) {
		fprintf(stderr, "Error parsing the formula!\n");
		return 1;	
	}
	
#ifdef DEBUG
	HL_MODEL_printFormula(stderr, theModel, theFormula, wvarTable);
	fprintf(stderr, "\n");
#endif

#ifdef DEBUG
	fprintf(stderr, "Model checking...\n");
#endif
	
	HL_FORMULA_modelChecking(theFormula, useMCLite);

#ifdef DEBUG
	fprintf(stderr, "The formula holds in following world(s):\n  ");
#endif
	HL_MODEL_printMCResults(stdout, theModel, theFormula);
	printf("\n");
	
	if(printStats)
		HL_FORMULA_printMCStats(stdout, useMCLite);
#ifdef DEBUG
	fprintf(stderr, "Cleaning up...\n");
#endif
	SYMBOL_TABLE_delete(wvarTable);
	HL_FORMULA_delete(theFormula);
	HL_MODEL_delete(theModel);

	return 0;
}
